# coding=utf-8
# --------------------------------------------------------------------------------------------------------------------
# Licensed under the MIT LICENSE. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------------------------------
"""Autocompounding and standard rewards issuance logic."""
from bancor_research.bancor_simulator.v3.spec.state import *


def calc_pool_token_amt_to_burn(
    staked_amt: Decimal,
    token_amt_to_distribute: Decimal,
    prev_token_amt_distributed: Decimal,
    erc20contracts_bntkn: Decimal,
    program_wallet_bntkn: Decimal,
) -> Decimal:
    """
    Return the token amt to burn for a given exponential distribution reward program.
    """
    a = staked_amt
    b = token_amt_to_distribute - prev_token_amt_distributed
    c = erc20contracts_bntkn
    d = program_wallet_bntkn
    return b * c**2 / (c * (a + b) - a * d)


def calc_flat_token_amt_to_distribute(
    state: State,
    tkn_name: str,
    time_elapsed: Decimal,
    total_rewards: Decimal,
) -> Decimal:
    """
    Return the token amt to distribute for a given flat distribution reward program.
    """
    flat_distribution_rate_per_second = get_flat_distribution_rate_per_second(
        state, tkn_name
    )
    return min(time_elapsed * flat_distribution_rate_per_second, total_rewards)


def calc_exp_token_amt_to_distribute(
    state: State, tkn_name: str, total_rewards: Decimal, time_elapsed: Decimal
) -> Decimal:
    """
    Return the token amt to distribute for a given exponential distribution reward program.
    """
    half_life = get_half_life(state, tkn_name)
    return total_rewards * (1 - (2 ** (-time_elapsed / half_life)))


def process_ac_rewards_program(state: State, tkn_name: str, timestamp: int) -> State:
    """
    Processes autocompounding rewards program for a given pool.
    """
    assert (
        tkn_name in state.autocompounding_reward_programs
    ), f"{tkn_name} not found in autocompounding_programs"

    staked_amt = get_staked_balance(state, tkn_name)
    erc20contracts_bntkn = get_pooltoken_balance(state, tkn_name)
    total_rewards = get_total_rewards(state, tkn_name)
    prev_token_amt_distributed = get_prev_token_amt_distributed(state, tkn_name)
    distribution_type = get_distribution_type(state, tkn_name)
    start_time = get_autocompounding_start_time(state, tkn_name)
    time_elapsed = Decimal(timestamp) - Decimal(start_time)

    if distribution_type == "flat":

        token_amt_to_distribute = calc_flat_token_amt_to_distribute(
            state=state,
            tkn_name=tkn_name,
            time_elapsed=time_elapsed,
            total_rewards=total_rewards,
        )
    else:

        token_amt_to_distribute = calc_exp_token_amt_to_distribute(
            state=state,
            tkn_name=tkn_name,
            total_rewards=total_rewards,
            time_elapsed=time_elapsed,
        )

    if tkn_name == "bnt":
        program_wallet_bntkn = get_protocol_wallet_balance(state, tkn_name)
    else:
        program_wallet_bntkn = get_external_rewards_vault_balance(state, tkn_name)

    pool_token_amt_to_burn = calc_pool_token_amt_to_burn(
        staked_amt=staked_amt,
        token_amt_to_distribute=token_amt_to_distribute,
        prev_token_amt_distributed=prev_token_amt_distributed,
        erc20contracts_bntkn=erc20contracts_bntkn,
        program_wallet_bntkn=program_wallet_bntkn,
    )

    # sanity check
    if pool_token_amt_to_burn >= erc20contracts_bntkn / 2:
        return terminate_ac_rewards_program(state, tkn_name, timestamp)

    if pool_token_amt_to_burn > 0:
        state.decrease_pooltoken_balance(tkn_name, pool_token_amt_to_burn)
        state.set_prev_token_amt_distributed(tkn_name, token_amt_to_distribute)
        remaining_rewards = total_rewards - token_amt_to_distribute
        state.set_autocompounding_remaining_rewards(tkn_name, remaining_rewards)
        if tkn_name == "bnt":
            state.decrease_protocol_wallet_balance(tkn_name, pool_token_amt_to_burn)
        else:
            state.decrease_external_rewards_vault_balance(
                tkn_name, pool_token_amt_to_burn
            )

    return state


def terminate_ac_rewards_program(state: State, tkn_name: str, timestamp: int) -> State:
    """
    Terminates autocompounding rewards program for a given pool.
    """
    del state.autocompounding_reward_programs[tkn_name]
    return state


"""
 * ------------------------------------------------------
 *
 * Everything below here refers to Standard Rewards
 *
 * ------------------------------------------------------
"""


def create_standard_reward_program(
    state: State,
    tkn_name: str,
    total_rewards: Decimal,
    start_time: int,
    end_time: int,
    timestamp: int,
) -> Tuple[State, int]:
    """
    Create a standard rewards program.
    """

    if not timestamp <= start_time and start_time < end_time:
        raise ValueError("standard program time constraints not met")

    if tkn_name in state.active_standard_programs:
        raise ValueError("standard program already exists")

    id = state.next_standard_program_id()

    reward_rate = total_rewards / (end_time - start_time)
    remaining_rewards = reward_rate * (end_time - start_time)

    state.standard_reward_programs[id] = StandardProgram(
        id=id,
        tkn_name=tkn_name,
        is_active=True,
        last_update_time=start_time,
        start_time=start_time,
        end_time=end_time,
        reward_rate=reward_rate,
    )

    state.set_standard_remaining_rewards(id, remaining_rewards)

    return state, id


def terminate_standard_program(state: State, id: int, timestamp: int) -> State:
    """
    Terminates the standard rewards program.
    """
    state.set_standard_program_is_active(id, False)
    state.set_standard_program_end_time(id, timestamp)
    return state


def snapshot_standard_rewards(
    state: State, id: int, timestamp: int, user_name: str
) -> State:
    """
    Snapshot of the existing standard rewards before modifying the staked_reward_amt balance.
    """
    state.map_user_standard_program(user_name, id)

    new_reward_per_token = calc_standard_reward_per_token(state, id, timestamp)
    state.set_standard_rewards_per_token(id, new_reward_per_token)

    end_time = get_standard_reward_end_time(state, id)
    last_update_time = get_standard_reward_last_update_time(state, id)
    state.set_standard_rewards_last_update_time(
        id, max(min(timestamp, end_time), last_update_time)
    )

    provider_staked_amount = get_user_pending_rewards_staked_balance(
        state, id, user_name
    )
    provider_pending_rewards = get_user_pending_standard_rewards(state, id, user_name)
    reward_per_token_paid = get_user_reward_per_token_paid(state, id, user_name)

    new_pending_rewards = calc_user_pending_standard_rewards(
        provider_pending_rewards,
        provider_staked_amount,
        new_reward_per_token,
        reward_per_token_paid,
    )

    state.set_provider_pending_standard_rewards(user_name, id, new_pending_rewards)
    state.set_provider_reward_per_token_paid(user_name, id, new_reward_per_token)
    return state


def calc_standard_reward_per_token(state: State, id: int, timestamp: int) -> Decimal:
    """
    Calculates the reward per token for the given program id.
    """
    total_staked = get_total_standard_rewards_staked(state, id)
    reward_per_token = get_standard_reward_per_token(state, id)

    if total_staked == 0:
        return reward_per_token

    start_time = get_standard_reward_start_time(state, id)
    end_time = get_standard_reward_end_time(state, id)
    last_update_time = get_standard_reward_last_update_time(state, id)
    reward_rate = get_standard_reward_rate(state, id)

    staking_end_time = min(timestamp, end_time)
    staking_start_time = max(start_time, last_update_time)

    return (
        reward_per_token
        + (staking_end_time - staking_start_time) * reward_rate / total_staked
    )


def calc_user_pending_standard_rewards(
    provider_pending_rewards: Decimal,
    provider_staked_amount: Decimal,
    updated_reward_per_token: Decimal,
    reward_per_token_paid: Decimal,
) -> Decimal:
    """
    Calculates the standard rewards pending for a given user
    """
    return provider_pending_rewards + provider_staked_amount * (
        updated_reward_per_token - reward_per_token_paid
    )


def calc_standard_rewards_remaining(state: State, id: int, timestamp: int) -> Decimal:
    """
    Calculates the standard rewards remaining for a given program id.
    """
    reward_rate = get_standard_reward_rate(state, id)
    end_time = get_standard_reward_end_time(state, id)
    return reward_rate * max(end_time - timestamp, 0)


def join_standard_reward_program(
    state: State, user_name: str, id: int, pool_token_amount: Decimal, timestamp: int
) -> State:
    """
    Join a standard reward program with the given user and id.
    """
    pool_token_name = get_standard_reward_pool_token_name(state, id)
    state = snapshot_standard_rewards(state, id, timestamp, user_name)
    state.decrease_user_balance(user_name, pool_token_name, pool_token_amount)
    state.increase_user_standard_rewards_stakes(id, user_name, pool_token_amount)
    return state


def leave_standard_reward_program(
    state: State, user_name: str, id: int, pool_token_amount: Decimal, timestamp: int
) -> State:
    """
    Leave a standard reward program for the given user and program id.
    """
    pool_token_name = get_standard_reward_pool_token_name(state, id)
    state = snapshot_standard_rewards(state, id, timestamp, user_name)
    state.decrease_user_standard_rewards_stakes(id, user_name, pool_token_amount)
    state.increase_user_balance(user_name, pool_token_name, pool_token_amount)
    return state


def claim_standard_rewards(
    state: State,
    user_name: str,
    ids: list,
    timestamp: int,
) -> State:
    """
    Claim standard rewards for the given user.
    """
    for id in ids:

        state = snapshot_standard_rewards(state, id, timestamp, user_name)
        reward_amt = get_user_pending_standard_rewards(state, id, user_name)
        remaining_rewards = get_standard_remaining_rewards(state, id)
        assert remaining_rewards >= reward_amt
        state.set_standard_remaining_rewards(id, remaining_rewards - reward_amt)
        state.set_user_pending_standard_rewards(user_name, id, Decimal(0))
        state.increase_user_balance(user_name, "bnt", reward_amt)

    return state



from bancor_research.bancor_emulator.solidity import uint, uint8, mapping
from bancor_research.bancor_emulator.utils import contract

from bancor_research.bancor_emulator.PoolToken import PoolToken

'''
 * @dev Pool Token Factory contract
'''
class PoolTokenFactory(contract):
    POOL_TOKEN_SYMBOL_PREFIX = "bn";
    POOL_TOKEN_NAME_PREFIX = "Bancor";
    POOL_TOKEN_NAME_SUFFIX = "Pool Token";

    def __init__(self) -> None:
        contract.__init__(self)

    '''
     * @dev fully initializes the contract and its parents
    '''
    def initialize(self) -> None:
        self.__PoolTokenFactory_init();

    # solhint-disable func-name-mixedcase

    '''
     * @dev initializes the contract and its parents
    '''
    def __PoolTokenFactory_init(self) -> None:
        self.__PoolTokenFactory_init_unchained();

    '''
     * @dev performs contract-specific initialization
    '''
    def __PoolTokenFactory_init_unchained(self) -> None:
        # a mapping between tokens and custom symbol overrides (only needed for tokens with malformed symbol property)
        self._tokenSymbolOverrides = mapping(lambda: str());

        # a mapping between tokens and custom token overrides (only needed for tokens with malformed decimals property)
        self._tokenDecimalsOverrides = mapping(lambda: uint8());

    # solhint-enable func-name-mixedcase

    '''
     * @inheritdoc Upgradeable
    '''
    def version(self) -> (int):
        return 1;

    '''
     * @inheritdoc IPoolTokenFactory
    '''
    def tokenSymbolOverride(self, token) -> (str):
        return self._tokenSymbolOverrides[token];

    '''
     * @dev sets the custom symbol override for a given reserve token
     *
     * requirements:
     *
     * - the caller must be the admin of the contract
    '''
    def setTokenSymbolOverride(self, token, symbol) -> None:
        self._tokenSymbolOverrides[token] = symbol;

    '''
     * @inheritdoc IPoolTokenFactory
    '''
    def tokenDecimalsOverride(self, token) -> (uint):
        return self._tokenDecimalsOverrides[token];

    '''
     * @dev sets the decimals override for a given reserve token
     *
     * requirements:
     *
     * - the caller must be the admin of the contract
    '''
    def setTokenDecimalsOverride(self, token, decimals) -> None:
        self._tokenDecimalsOverrides[token] = decimals;

    '''
     * @inheritdoc IPoolTokenFactory
    '''
    def createPoolToken(self, token) -> (PoolToken):
        customSymbol = self._tokenSymbolOverrides[token];
        tokenSymbol = customSymbol if len(customSymbol) != 0 else token.symbol();

        customDecimals = self._tokenDecimalsOverrides[token];
        tokenDecimals = customDecimals if customDecimals != 0 else token.decimals();

        symbol = "".join([self.POOL_TOKEN_SYMBOL_PREFIX, tokenSymbol]);
        name = "".join([self.POOL_TOKEN_NAME_PREFIX, " ", tokenSymbol, " ", self.POOL_TOKEN_NAME_SUFFIX]);

        newPoolToken = PoolToken(name, symbol, tokenDecimals, token);

        # make sure to transfer the ownership to the caller
        newPoolToken.transferOwnership(self.msg_sender);

        return newPoolToken;



import decimal
from decimal import Decimal
import pandas as pd
from bancor_research.bancor_simulator.v3.spec import *
import random
from statistics import mean


class MonteCarloGenerator(object):
    """
    Generates Monte Carlo scenarios.
    """

    def __init__(
            self,
            whitelisted_tokens: dict,
            price_feed: pd.DataFrame,
            user_initial_balances: pd.DataFrame,
            simulation_actions_count: int,
            num_timesteps: int,
            num_simulation_days: int,
            pool_freq_dist: dict,
            action_freq_dist: dict,
            deposit_mean: float,
            trade_mean: float,
            withdraw_mean: float,
            cooldown_time: int = 0,
            bnt_min_liquidity: Any = 10000,
    ):

        # all users/agents use a single BancorDapp instance
        v3 = BancorDapp(
            cooldown_time=cooldown_time,
            whitelisted_tokens=whitelisted_tokens,
            bnt_min_liquidity=bnt_min_liquidity,
            price_feeds=price_feed,
        )

        # set the initial balances for each user.
        for user_id in user_initial_balances["user_id"].unique():
            v3.create_user(user_id)

            user_balances = user_initial_balances[
                user_initial_balances["user_id"] == user_id
                ]
            for tkn_name in user_balances["poolSymbol"].unique():

                user_balance = user_balances[user_balances["poolSymbol"] == tkn_name][
                    "tokenAmount_real_usd"
                ].values[0]
                v3.set_user_balance(
                    user_name=user_id,
                    tkn_name=tkn_name,
                    tkn_amt=user_balance,
                    timestamp=0,
                )

                pooltkn_name = get_pooltoken_name(tkn_name)
                if pooltkn_name not in v3.global_state.users[user_id].wallet:
                    v3.global_state.users[user_id].wallet[pooltkn_name] = Token(
                        balance=Decimal("0")
                    )

        self.protocol = v3
        self.random = random
        self.logger = []
        self.timestamp = 0
        self.simulation_actions_count = simulation_actions_count
        self.whitelisted_tokens = whitelisted_tokens
        self.daily_trade_volume = 0
        self.latest_amt = 0
        self.latest_tkn_name = None
        self.rolling_trade_fees = {}
        self.total_fees_earned = {}
        self.action_freq_dist = action_freq_dist
        self.num_timesteps = num_timesteps
        self.slippage_profile = {}
        self.user_initial_balances = user_initial_balances
        self.iloss_tracker = {}
        self.iloss_realized = {}
        self.total_fees_earned = {}
        self.num_simulation_days = num_simulation_days
        self.slippage_pearson_correlation = 0
        self.deposit_mean = deposit_mean
        self.trade_mean = trade_mean
        self.withdraw_mean = withdraw_mean
        random_tkn_names = []

        # create a list of tokens which occur at the desired frequency
        for tkn in pool_freq_dist:
            self.total_fees_earned[tkn] = [0]
            self.rolling_trade_fees[tkn] = []
            self.iloss_realized[tkn] = [0]
            freq = int(round(float(pool_freq_dist[tkn] * simulation_actions_count), 0))
            for i in range(freq):
                random_tkn_names.append(tkn)

        # randomly shuffle the list of tokens that we will select from
        random.seed(1)
        for i in range(50):
            random.shuffle(random_tkn_names)
        self.pool_freq_dist_list = random_tkn_names

    def get_random_deposit_amt(self, amt: Decimal = None) -> Decimal:
        if amt is None:
            amt = Decimal(self.deposit_mean)
        max_amt, min_amt = amt * Decimal("100.0"), amt / Decimal("100.0")
        return Decimal(self.random.uniform(float(min_amt), float(max_amt)))

    def get_random_trade_amt(self, amt: Decimal = None) -> Decimal:
        if amt is None:
            amt = Decimal(self.trade_mean)
        max_amt, min_amt = amt * Decimal("100.0"), amt / Decimal("100.0")
        return Decimal(self.random.uniform(float(min_amt), float(max_amt)))

    def get_random_withdraw_amt(self, amt: Decimal = None) -> Decimal:
        if amt is None:
            amt = Decimal(self.withdraw_mean)
        max_amt, min_amt = amt * Decimal("100.0"), amt / Decimal("100.0")
        return Decimal(self.random.uniform(float(min_amt), float(max_amt)))

    def handle_trade_fees(self, target_tkn: str, source_tkn: str):
        """
        Collects data on the trade fees earned during the most recent trade action.
        """
        state = self.protocol.global_state

        if source_tkn == "bnt":
            if target_tkn in state.rolling_trade_fees:
                fees_earned = state.rolling_trade_fees[target_tkn][-1]
                self.rolling_trade_fees[target_tkn].append(fees_earned)
                self.total_fees_earned[target_tkn].append(
                    float(self.total_fees_earned[target_tkn][-1]) + float(fees_earned)
                )

        elif target_tkn == "bnt":
            if source_tkn in state.rolling_trade_fees:
                fees_earned = state.rolling_trade_fees[source_tkn][-1]
                self.rolling_trade_fees[source_tkn].append(fees_earned)
                self.total_fees_earned[source_tkn].append(
                    float(self.total_fees_earned[source_tkn][-1]) + float(fees_earned)
                )

        elif target_tkn != "bnt" and source_tkn != "bnt":
            if source_tkn in state.rolling_trade_fees:
                fees_earned_1 = state.rolling_trade_fees[source_tkn][-1]
                self.rolling_trade_fees[source_tkn].append(fees_earned_1)
                self.total_fees_earned[source_tkn].append(
                    float(self.total_fees_earned[source_tkn][-1]) + float(fees_earned_1)
                )

            if target_tkn in state.rolling_trade_fees:
                fees_earned_2 = state.rolling_trade_fees[target_tkn][-1]
                self.rolling_trade_fees[target_tkn].append(fees_earned_2)
                self.total_fees_earned[target_tkn].append(
                    float(self.total_fees_earned[target_tkn][-1]) + float(fees_earned_2)
                )

    def force_moving_average(
            self, state: State, tkn_name: str, user_name: str, timestamp: int
    ):
        """
        Analyze the state of a pool and creates a swap that drives the ema and the spot price together.
        """
        user_tkn = get_user_balance(state, user_name, tkn_name)
        user_bnt = get_user_balance(state, user_name, "bnt")
        if get_is_trading_enabled(state, tkn_name):
            (
                tkn_amt,
                source_token,
                target_token,
                user_capability,
            ) = process_force_moving_average(tkn_name, user_tkn, user_bnt)
            if user_capability:
                self.protocol.v3.trade(
                    tkn_amt, source_token, target_token, user_name, timestamp
                )
                self.latest_tkn_name = source_token + "_" + target_token
                self.latest_amt = tkn_amt

    def get_slippage(self, amt, tkn_trading_liquidity, tkn_name) -> Decimal:
        """
        Calculates the slippage for a given trade, and enforces historical adjustments.
        """
        slippage_perc = amt / (amt + tkn_trading_liquidity)

        # derived from real-world data
        historical_max_slippage = self.slippage_profile[tkn_name]["max"]
        historical_min_slippage = self.slippage_profile[tkn_name]["min"]

        if slippage_perc > historical_max_slippage:

            # if slippage is equal to or greater than the max., impact equals historical_max_slippage
            slippage_perc = historical_max_slippage

        elif slippage_perc < historical_min_slippage:

            # if slippage is equal to or less than the min., impact equals historical_min_slippage
            slippage_perc = historical_min_slippage

        return slippage_perc

    def perform_random_trade(self):
        """
        Performs a random trade on the server.
        """
        state = self.protocol.global_state
        timestamp = self.timestamp
        user_name = "global user"
        source_tkn, target_tkn = self.get_random_tkn_names(state)
        user_source_before = get_user_balance(state, user_name, source_tkn)
        swap_amt = self.get_random_trade_amt()

        if user_source_before > swap_amt:
            amt = swap_amt
        else:
            amt = user_source_before / Decimal("1000.0")

        if amt > 0:
            self.protocol.trade(
                tkn_amt=str(amt),
                source_token=source_tkn,
                target_token=target_tkn,
                user_name=user_name,
                timestamp=timestamp,
            )
            self.latest_tkn_name = source_tkn + "_" + target_tkn
            self.latest_amt = amt
            self.handle_trade_fees(target_tkn, source_tkn)

    def perform_random_arbitrage_trade(self):
        """
        Performs a random arbitrage trade.
        """
        state = self.protocol.global_state
        user_name = "global user"
        tkn_name, target_tkn = self.get_random_tkn_names(state)
        timestamp = self.timestamp
        tkn_price, bnt_price = get_prices(state, tkn_name)
        bnt_trading_liquidity = get_bnt_trading_liquidity(state, tkn_name)
        tkn_trading_liquidity = get_tkn_trading_liquidity(state, tkn_name)
        trading_fee = get_trading_fee(state, tkn_name)
        user_tkn = get_user_balance(state, user_name, tkn_name)
        user_bnt = get_user_balance(state, user_name, "bnt")
        trade_amt = Decimal("0")
        source_token = tkn_name
        target_token = target_tkn
        x = process_arbitrage_trade(
            tkn_name,
            tkn_price,
            bnt_price,
            bnt_trading_liquidity,
            tkn_trading_liquidity,
            trading_fee,
            user_tkn,
            user_bnt,
        )
        if x is not None:
            (
                trade_amt,
                source_token,
                target_token,
                user_capability,
            ) = x
            if user_capability:
                if trade_amt > 0:
                    self.protocol.trade(
                        str(trade_amt), source_token, target_token, user_name, timestamp
                    )

            source_tkn = source_token
            target_tkn = target_token
            self.handle_trade_fees(target_tkn, source_tkn)

        self.latest_tkn_name = source_token + "_" + target_token
        self.latest_amt = trade_amt

    def get_random_tkn_names(self, state: State) -> Tuple[str, str]:
        source_tkn, target_tkn = "None", "None"
        while source_tkn == target_tkn:
            source_tkn, target_tkn = self.random.sample(self.pool_freq_dist_list, 2)
        if source_tkn == target_tkn:
            target_tkn = self.random.sample(
                self.pool_freq_dist_list.remove(source_tkn), 1
            )
        return source_tkn, target_tkn

    def get_random_withdrawal_amt(self, tkn_name: str) -> Decimal:
        user_balance = get_user_balance(
            self.protocol.global_state, self.user_name, tkn_name
        )
        return user_balance * Decimal(self.random.uniform(float(0.0001), float(0.5)))

    def get_random_cooldown_amt(self, user_bntkn_amt: Decimal = None) -> Decimal:

        if user_bntkn_amt is None:
            user_bntkn_amt = Decimal(self.withdraw_mean)

        max_amt, min_amt = user_bntkn_amt * Decimal("10.0"), user_bntkn_amt * Decimal(
            "0.00001"
        )
        return Decimal(self.random.uniform(float(min_amt), float(max_amt)))

    def is_protocol_bnbnt_healthy(
            self, protocol_bnbnt: Decimal, bnbnt_supply: Decimal
    ) -> bool:
        """
        Checks if the protocol owned bnbnt is at a healthy level (greater than 50%)
        """
        return protocol_bnbnt / bnbnt_supply > Decimal("0.5")

    def get_deposit_payload(
            self, state: State, user_name: str
    ) -> Tuple[str, str, Decimal, Decimal, Decimal, Decimal, Decimal]:
        """
        Gets the input data required for a deposit.
        """
        tkn_name, target_tkn = self.get_random_tkn_names(state)
        user_tkn = get_user_balance(state, user_name, tkn_name)
        user_bnt = get_user_balance(state, user_name, "bnt")
        bnbnt_supply = get_pooltoken_balance(state, "bnt")
        protocol_bnbnt = get_protocol_wallet_balance(state, "bnt")
        bnbnt_rate = get_bnbnt_rate(state)
        return (
            user_name,
            tkn_name,
            user_tkn,
            user_bnt,
            bnbnt_supply,
            protocol_bnbnt,
            bnbnt_rate,
        )

    def perform_random_enable_trading(self):
        """
        Begins a random withdrawal.
        """
        state = self.protocol.global_state
        timestamp = state.timestamp
        tkn_name, target_tkn = self.get_random_tkn_names(state)
        if not get_is_trading_enabled(state, tkn_name):
            self.protocol.enable_trading(tkn_name=tkn_name, timestamp=timestamp)

    def perform_random_withdrawal(self):
        """
        Begins a random withdrawal.
        """
        state = self.protocol.global_state
        timestamp = self.timestamp
        user_name = self.random.choice(
            [n for n in state.usernames if n not in ["protocol", "global user"]]
        )
        tkn_name, target_tkn = self.get_random_tkn_names(state)
        user_bntkn_amt = get_user_balance(state, user_name, f"bn{tkn_name}")
        bntkn_rate = compute_bntkn_rate(state, tkn_name)
        if user_bntkn_amt > 0 and bntkn_rate > 0:
            bntkn_amt = user_bntkn_amt
            withdraw_value = bntkn_amt / bntkn_rate
            id_number = self.protocol.begin_cooldown_by_rtkn(
                tkn_amt=str(withdraw_value),
                tkn_name=tkn_name,
                user_name=user_name,
                timestamp=timestamp,
            )
            self.protocol.withdraw(
                user_name=user_name, id_number=id_number, timestamp=timestamp
            )
            tkn_price_initial = float(self.iloss_tracker[user_name]["tkn_price"])
            bnt_price_initial = float(self.iloss_tracker[user_name]["bnt_price"])
            tkn_price_final = float(get_tkn_price(state, tkn_name))
            bnt_price_final = float(get_tkn_price(state, "bnt"))

            var_A = (
                            (tkn_price_final - tkn_price_initial) / tkn_price_initial
                    ) * 100
            var_B = (
                            (bnt_price_final - bnt_price_initial) / bnt_price_initial
                    ) * 100

            iloss_realized = float(self.iloss(var_A, var_B)) * float(withdraw_value)
            self.iloss_tracker[user_name] = {
                "tkn_name": tkn_name,
                "tkn_price": tkn_price_final,
                "bnt_price": bnt_price_final,
                "timestamp": timestamp,
                "iloss_realized": iloss_realized,
            }

            self.iloss_realized[tkn_name].append(
                float(self.iloss_realized[tkn_name][-1]) + iloss_realized
            )
            self.latest_tkn_name = tkn_name
            self.latest_amt = withdraw_value
        return self

    def iloss(self, var_A: float, var_B: float) -> float:
        """Returns the impermanent loss value.

        Args:
            var_A: Asset A % variation
            var_B: Asset B % variation
        Returns:
            IL
        """
        price_ratio = (float(var_A) / 100 + 1) / (float(var_B) / 100 + 1)
        il = 2 * (price_ratio ** 0.5 / (1 + price_ratio)) - 1
        return il

    def perform_random_deposit(self):
        """
        Performs a random deposit action
        """

        state = self.protocol.global_state
        timestamp = self.timestamp
        user_name = f"user_{timestamp}"
        self.protocol.create_user(user_name)
        user_balances = self.user_initial_balances[
            self.user_initial_balances["user_id"] == "global user"
            ]
        for tkn_name in user_balances["poolSymbol"].unique():
            user_balance = user_balances[user_balances["poolSymbol"] == tkn_name][
                "tokenAmount_real_usd"
            ].values[0]
            self.protocol.set_user_balance(
                user_name=user_name,
                tkn_name=tkn_name,
                tkn_amt=user_balance,
                timestamp=timestamp,
            )
            pooltkn_name = get_pooltoken_name(tkn_name)
            if pooltkn_name not in self.protocol.global_state.users[user_name].wallet:
                self.protocol.global_state.users[user_name].wallet[
                    pooltkn_name
                ] = Token(balance=Decimal("0"))

        state = self.protocol.global_state

        (
            user_name,
            tkn_name,
            user_tkn,
            user_bnt,
            bnbnt_supply,
            protocol_bnbnt,
            bnbnt_rate,
        ) = self.get_deposit_payload(state, user_name)
        deposit_amt = None
        if tkn_name != "bnt":
            deposit_amt = self.get_random_deposit_amt()
            if 0 < deposit_amt < user_tkn:
                self.protocol.deposit(
                    tkn_name,
                    str(deposit_amt),
                    user_name,
                    timestamp,
                )
        elif tkn_name == "bnt":
            if bnbnt_supply > 0 and self.is_protocol_bnbnt_healthy(
                    protocol_bnbnt, bnbnt_supply
            ):
                deposit_amt = self.get_random_deposit_amt()
                if 0 < deposit_amt < user_bnt:
                    self.protocol.deposit(
                        tkn_name, str(deposit_amt), user_name, timestamp
                    )

        self.latest_tkn_name = tkn_name
        self.latest_amt = deposit_amt
        self.iloss_tracker[user_name] = {
            "tkn_name": tkn_name,
            "tkn_price": get_tkn_price(state, tkn_name),
            "bnt_price": get_tkn_price(state, "bnt"),
            "timestamp": timestamp,
            "iloss_realized": 0.0,
        }

    def update_trading_liquidity(self, n_events: int, constant_multiplier: int = 520):
        """
        Updates the state of the appropriate pool, and the protocol holdings, as required.
        """

        state = self.protocol.global_state
        for tkn_name in list(state.whitelisted_tokens):
            if len(self.rolling_trade_fees[tkn_name]) > 0:

                # Calculate the new value for bnt_funding_limit based on the rolling avg trade fees.
                new_bnt_funding_limit = (
                        mean(self.rolling_trade_fees[tkn_name][-n_events:])
                        * constant_multiplier
                )

                # Convert the tkn units into BNT
                bnt_per_tkn = get_bnt_per_tkn(state, tkn_name)
                new_bnt_funding_limit = new_bnt_funding_limit * bnt_per_tkn
                updated_bnt_trading_liquidity = Decimal(new_bnt_funding_limit)

                # Get the current trading liquidity
                bnt_trading_liquidity = get_bnt_trading_liquidity(state, tkn_name)
                tkn_trading_liquidity = get_tkn_trading_liquidity(state, tkn_name)

                # Calculate the change in trading liquidity
                bnt_delta = bnt_trading_liquidity - updated_bnt_trading_liquidity

                # If the price is stable, perform the update
                if get_is_price_stable(state, tkn_name) and new_bnt_funding_limit > 0:

                    if bnt_delta > 0:
                        # If we are reducing the trading liquidity, make the necessary adjustments
                        state.decrease_staked_balance(tkn_name, bnt_delta)
                        state.decrease_master_vault_balance(tkn_name, bnt_delta)
                        state.decrease_pooltoken_balance(tkn_name, bnt_delta)
                        state.decrease_protocol_wallet_balance(tkn_name, bnt_delta)
                        updated_tkn_trading_liquidity = max(
                            tkn_trading_liquidity - bnt_delta, 0
                        )
                        state.set_tkn_trading_liquidity(
                            tkn_name, updated_tkn_trading_liquidity
                        )
                        state.set_bnt_trading_liquidity(
                            tkn_name, updated_bnt_trading_liquidity
                        )
                        state.set_bnt_funding_amt(
                            tkn_name, updated_bnt_trading_liquidity
                        )

                    # In order to avoid automatic pool shutdown, we change this parameter dynamically also
                    state.tokens[tkn_name].bnt_min_liquidity = Decimal(
                        updated_bnt_trading_liquidity
                    ) * Decimal("0.5")

                    state.set_bnt_funding_limit(tkn_name, updated_bnt_trading_liquidity)

                    if check_pool_shutdown(state, tkn_name):
                        state = shutdown_pool(state, tkn_name)

                    self.protocol.set_state(state)

    def run(
            self,
            transact,
            mean_events_per_day=None,
            num_timestamps_per_day=None,
            n_rolling_days=None,
            constant_multiplier=None,
            is_proposal=False,
    ):
        """
        Executes the simulation run and returns the collected data.
        """

        day_num = 0
        for _ in range(self.num_simulation_days):
            self.timestamp += 1
            day_num += 1
            transact(self)
            if is_proposal:
                if day_num > n_rolling_days:
                    n_events = int(round(mean_events_per_day * n_rolling_days, 0))
                    backup_state = self.protocol.global_state.copy()
                    try:
                        try:
                            self.update_trading_liquidity(n_events, constant_multiplier)
                        except KeyError:
                            print(f'The price feed is too short to complete the run. Early stopping will result.')
                            self.protocol.global_state = backup_state
                            break
                    except AssertionError as e:
                        print(f'Proposal trading liquidity update failed, reverting state. {e}')
                        self.protocol.global_state = backup_state

        return pd.concat(self.logger)



import re

from setuptools import find_packages, setup

with open("bancor_research/__init__.py") as fd:
    version = re.search(
        r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', fd.read(), re.MULTILINE
    ).group(1)

with open("README.md", "r", encoding="utf8") as fh:
    long_description = fh.read()

extras_require = {}
extras_require['complete'] = sorted(set(sum(extras_require.values(), [])))

setup(
    name='bancor_research',
    version=version,
    author='Bancor Network',
    author_email='mike@bancor.network',
    description='Bancor v3 python spec, simulator, and emulator library. This is an open-source python package developed by Bancor Research. It is meant to assist in the design, testing, and validating of Bancor v3 behavior.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/bancorprotocol/research',
    install_requires=open('requirements.txt').readlines(),
    extras_require=extras_require,
    tests_require=open('test-requirements.txt').readlines(),
    packages=find_packages(),
    include_package_data=True,
    entry_points={
        'console_scripts': [
          'bancor_simulator = bancor_research.bancor_simulator.__main__:cli'
        ]
    }
)



from bancor_research.bancor_emulator.utils import contract

# @title The Token Governance contract is used to govern a mintable ERC20 token by restricting its launch-time initial
# administrative privileges.
class TokenGovernance(contract):
    # @dev Initializes the contract.
    #
    # @param mintableToken The address of the mintable ERC20 token.
    def __init__(self, mintableToken):
        contract.__init__(self)

        self._token = mintableToken;

    # @dev Mints new tokens. Only allowed by the MINTER role.
    #
    # @param to Account to receive the new amount.
    # @param amount Amount to increase the supply by.
    #
    def mint(self, to, amount):
        self._token.issue(to, amount);

    # @dev Burns tokens from the caller.
    #
    # @param amount Amount to decrease the supply by.
    #
    def burn(self, amount):
        self._token.destroy(self.msg_sender, amount);

    def token(self):
        return self._token;



import inspect

def library(globalVars, classHandle):
    for varName in vars(classHandle):
        if not varName.startswith('__'):
            setattr(classHandle, varName, globalVars[varName])

def using(libraryHandle, classHandle):
    for varName in vars(libraryHandle):
        if not varName.startswith('__'):
            setattr(classHandle, varName, getattr(libraryHandle, varName))

def parse(cast, var, attr):
    if type(var) is dict:
        return cast(var[attr])
    if hasattr(var, attr):
        return cast(getattr(var, attr))
    if var is not None:
        return cast(var)
    return cast(0)

class contract:
    def __init__(self):
        self._msg_sender = None

    def connect(self, _msg_sender):
        self._msg_sender = _msg_sender
        return self

    @property
    def msg_sender(self):
        f_locals = inspect.currentframe().f_back.f_back.f_locals
        return f_locals['self'] if 'self' in f_locals and isinstance(f_locals['self'], contract) else self._msg_sender



from bancor_research.bancor_emulator.solidity import uint, uint256, mapping, address, revert
from bancor_research.bancor_emulator.utils import parse

from bancor_research.bancor_emulator.Math import Math
from bancor_research.bancor_emulator.Constants import PPM_RESOLUTION
from bancor_research.bancor_emulator.MathEx import MathEx
from bancor_research.bancor_emulator.Vault import Vault
from bancor_research.bancor_emulator.PoolToken import PoolToken as IPoolToken

'''
 * @dev BNT Pool contract
'''
class BNTPool(Vault):
    class InternalWithdrawalAmounts:
        def __init__(self, x = None) -> None:
            self.bntAmount = parse(uint256, x, 'bntAmount');
            self.withdrawalFeeAmount = parse(uint256, x, 'withdrawalFeeAmount');

    '''
     * @dev a "virtual" constructor that is only used to set immutable state variables
    '''
    def __init__(self,
        initNetwork,
        initBNTGovernance,
        initVBNTGovernance,
        initNetworkSettings,
        initMasterVault,
        initBNTPoolToken
    ) -> None:
        Vault.__init__(self, initBNTGovernance, initVBNTGovernance)

        self._network = initNetwork;
        self._networkSettings = initNetworkSettings;
        self._masterVault = initMasterVault;
        self._poolToken = initBNTPoolToken;

    '''
     * @dev fully initializes the contract and its parents
    '''
    def initialize(self) -> None:
        self.__BNTPool_init();

    # solhint-disable func-name-mixedcase

    '''
     * @dev initializes the contract and its parents
    '''
    def __BNTPool_init(self) -> None:
        self.__BNTPool_init_unchained();

    '''
     * @dev performs contract-specific initialization
    '''
    def __BNTPool_init_unchained(self) -> None:
        self._stakedBalance = uint256()
        self._currentPoolFunding = mapping(lambda: uint256())

    # solhint-enable func-name-mixedcase

    '''
     * @inheritdoc Upgradeable
    '''
    def version(self) -> (int):
        return 3;

    '''
     * @inheritdoc IBNTPool
    '''
    def poolToken(self) -> (IPoolToken):
        return self._poolToken;

    '''
     * @inheritdoc IBNTPool
    '''
    def stakedBalance(self) -> (uint):
        return self._stakedBalance.clone();

    '''
     * @inheritdoc IBNTPool
    '''
    def currentPoolFunding(self, pool) -> (uint):
        return self._currentPoolFunding[pool].clone();

    '''
     * @inheritdoc IBNTPool
    '''
    def availableFunding(self, pool) -> (uint):
        return MathEx.subMax0(self._networkSettings.poolFundingLimit(pool), self._currentPoolFunding[pool]);

    '''
     * @inheritdoc IBNTPool
    '''
    def poolTokenToUnderlying(self, poolTokenAmount) -> (uint):
        return self._poolTokenToUnderlying(poolTokenAmount);

    '''
     * @inheritdoc IBNTPool
    '''
    def underlyingToPoolToken(self, bntAmount) -> (uint):
        return self._underlyingToPoolToken(bntAmount);

    '''
     * @inheritdoc IBNTPool
    '''
    def poolTokenAmountToBurn(self, bntAmountToDistribute) -> (uint):
        if (bntAmountToDistribute == 0):
            return uint256(0);

        poolTokenSupply = self._poolToken.totalSupply();
        val = bntAmountToDistribute * poolTokenSupply;

        return MathEx.mulDivF(
            val,
            poolTokenSupply,
            val + self._stakedBalance * (poolTokenSupply - self._poolToken.balanceOf(address(self)))
        );

    '''
     * @inheritdoc IBNTPool
    '''
    def mint(self, recipient, bntAmount):
        self._bntGovernance.mint(recipient, bntAmount);

    '''
     * @inheritdoc IBNTPool
    '''
    def burnFromVault(self, bntAmount):
        self._masterVault.burn(address(self._bnt), bntAmount);

    '''
     * @inheritdoc IBNTPool
    '''
    def depositFor(self,
        contextId,
        provider,
        bntAmount,
        isMigrating,
        originalVBNTAmount
    ):
        # calculate the required pool token amount
        currentStakedBalance = self._stakedBalance;
        poolTokenTotalSupply = self._poolToken.totalSupply();
        if (poolTokenTotalSupply == 0 and currentStakedBalance > 0):
            revert("InvalidStakedBalance");

        poolTokenAmount = self._underlyingToPoolToken_(bntAmount, poolTokenTotalSupply, currentStakedBalance);

        # if the protocol doesn't have enough pool tokens, mint new ones
        poolTokenBalance = self._poolToken.balanceOf(address(self));
        if (poolTokenAmount > poolTokenBalance):
            newPoolTokenAmount = poolTokenAmount - poolTokenBalance;
            increaseStakedBalanceAmount = self._poolTokenToUnderlying_(
                newPoolTokenAmount,
                currentStakedBalance,
                poolTokenTotalSupply
            );

            # update the staked balance
            self._stakedBalance = currentStakedBalance + increaseStakedBalanceAmount;

            # mint pool tokens to the protocol
            self._poolToken.mint(address(self), newPoolTokenAmount);

        # transfer pool tokens from the protocol to the provider
        self._poolToken.transfer(provider, poolTokenAmount);

        # burn the previously received BNT
        self._bntGovernance.burn(bntAmount);

        vbntAmount = poolTokenAmount;

        # the provider should receive pool tokens and vBNT in equal amounts. since the provider might already have
        # some vBNT during migration, the contract only mints the delta between the full amount and the amount the
        # provider already has
        if (isMigrating):
            vbntAmount = MathEx.subMax0(vbntAmount, originalVBNTAmount);

        # mint vBNT to the provider
        if (vbntAmount > 0):
            self._vbntGovernance.mint(provider, vbntAmount);

        return poolTokenAmount;

    '''
     * @inheritdoc IBNTPool
    '''
    def withdraw(self,
        contextId,
        provider,
        poolTokenAmount,
        bntAmount
    ):
        # ensure that the provided amounts correspond to the state of the pool. Note the pool tokens should
        # have been already deposited back from the network
        underlyingAmount = self._poolTokenToUnderlying(poolTokenAmount);
        if (bntAmount > underlyingAmount):
            revert("InvalidParam");

        amounts = self._withdrawalAmounts(bntAmount);

        # burn the respective vBNT amount
        self._vbntGovernance.burn(poolTokenAmount);

        # mint BNT to the provider
        self._bntGovernance.mint(provider, amounts.bntAmount);

        return amounts.bntAmount;

    '''
     * @inheritdoc IBNTPool
    '''
    def withdrawalAmount(self, poolTokenAmount):
        return self._withdrawalAmounts(self._poolTokenToUnderlying(poolTokenAmount)).bntAmount;

    '''
     * @inheritdoc IBNTPool
    '''
    def requestFunding(self,
        contextId,
        pool,
        bntAmount
    ):
        currentFunding = self._currentPoolFunding[pool];
        fundingLimit = self._networkSettings.poolFundingLimit(pool);
        newFunding = currentFunding + bntAmount;

        # verify that the new funding amount doesn't exceed the limit
        if (newFunding > fundingLimit):
            revert("FundingLimitExceeded");

        # calculate the pool token amount to mint
        currentStakedBalance = self._stakedBalance;
        poolTokenAmount = uint256();
        poolTokenTotalSupply = self._poolToken.totalSupply();
        if (poolTokenTotalSupply == 0 and currentStakedBalance > 0):
            revert("InvalidStakedBalance");

        poolTokenAmount = self._underlyingToPoolToken_(bntAmount, poolTokenTotalSupply, currentStakedBalance);

        # update the staked balance
        newStakedBalance = currentStakedBalance + bntAmount;
        self._stakedBalance = newStakedBalance;

        # update the current funding amount
        self._currentPoolFunding[pool] = newFunding;

        # mint pool tokens to the protocol
        self._poolToken.mint(address(self), poolTokenAmount);

        # mint BNT to the vault
        self._bntGovernance.mint(address(self._masterVault), bntAmount);

    '''
     * @inheritdoc IBNTPool
    '''
    def renounceFunding(self,
        contextId,
        pool,
        bntAmount
    ):
        currentStakedBalance = self._stakedBalance;

        # calculate the final amount to deduct from the current pool funding
        currentFunding = self._currentPoolFunding[pool];
        reduceFundingAmount = Math.min(currentFunding, bntAmount);

        # calculate the amount of pool tokens to burn
        # note that the given amount can exceed the total available but the request shouldn't fail
        poolTokenTotalSupply = self._poolToken.totalSupply();
        poolTokenAmount = self._underlyingToPoolToken_(
            reduceFundingAmount,
            poolTokenTotalSupply,
            currentStakedBalance
        );

        # ensure the amount of pool tokens doesn't exceed the total available
        poolTokenAmount = Math.min(poolTokenAmount, self._poolToken.balanceOf(address(self)));

        # calculate the final amount to deduct from the staked balance
        reduceStakedBalanceAmount = self._poolTokenToUnderlying_(
            poolTokenAmount,
            currentStakedBalance,
            poolTokenTotalSupply
        );

        # update the current pool funding. Note that the given amount can exceed the funding amount but the
        # request shouldn't fail (and the funding amount cannot get negative)
        self._currentPoolFunding[pool] = currentFunding - reduceFundingAmount;

        # update the staked balance
        newStakedBalance = currentStakedBalance - reduceStakedBalanceAmount;
        self._stakedBalance = newStakedBalance;

        # burn pool tokens from the protocol
        self._poolToken.burn(poolTokenAmount);

        # burn all BNT from the master vault
        self._masterVault.burn(address(self._bnt), bntAmount);

    '''
     * @inheritdoc IBNTPool
    '''
    def onFeesCollected(self,
        pool,
        feeAmount,
        isTradeFee
    ):
        if (feeAmount == 0):
            return;

        # increase the staked balance by the given amount
        self._stakedBalance += feeAmount;

        if (isTradeFee):
            # increase the current funding for the specified pool by the given amount
            self._currentPoolFunding[pool] += feeAmount;

    '''
     * @dev converts the specified pool token amount to the underlying BNT amount
    '''
    def _poolTokenToUnderlying(self, poolTokenAmount) -> (uint):
        return self._poolTokenToUnderlying_(poolTokenAmount, self._stakedBalance, self._poolToken.totalSupply());

    '''
     * @dev converts the specified pool token amount to the underlying BNT amount
    '''
    def _poolTokenToUnderlying_(self,
        poolTokenAmount,
        currentStakedBalance,
        poolTokenTotalSupply
    ) -> (uint):
        # if no pool token supply exists yet, use a one-to-one pool token to BNT rate
        if (poolTokenTotalSupply == 0):
            return poolTokenAmount;

        return MathEx.mulDivF(poolTokenAmount, currentStakedBalance, poolTokenTotalSupply);

    '''
     * @dev converts the specified underlying BNT amount to pool token amount
    '''
    def _underlyingToPoolToken(self, bntAmount) -> (uint):
        return self._underlyingToPoolToken_(bntAmount, self._poolToken.totalSupply(), self._stakedBalance);

    '''
     * @dev converts the specified underlying BNT amount to pool token amount
    '''
    def _underlyingToPoolToken_(self,
        bntAmount,
        poolTokenTotalSupply,
        currentStakedBalance
    ) -> (uint):
        # if no pool token supply exists yet, use a one-to-one pool token to BNT rate
        if (poolTokenTotalSupply == 0):
            return bntAmount;

        return MathEx.mulDivC(bntAmount, poolTokenTotalSupply, currentStakedBalance);

    '''
     * @dev returns withdrawal amounts
    '''
    def _withdrawalAmounts(self, bntAmount) -> (InternalWithdrawalAmounts):
        # deduct the exit fee from BNT amount
        withdrawalFeeAmount = MathEx.mulDivF(bntAmount, self._networkSettings.withdrawalFeePPM(), PPM_RESOLUTION);

        bntAmount -= withdrawalFeeAmount;

        return self.InternalWithdrawalAmounts({ 'bntAmount': bntAmount, 'withdrawalFeeAmount': withdrawalFeeAmount });



'''
 * @dev Library for managing
 * https:#en.wikipedia.org/wiki/Set_(abstract_data_type)[sets] of primitive
 * types.
 *
 * Sets have the following properties:
 *
 * - Elements are added, removed, and checked for existence in constant time
 * (O(1)).
 * - Elements are enumerated in O(n). No guarantees are made on the ordering.
 *
 * ```
 * contract Example {
 *     # Add the library methods
 *     using EnumerableSet for EnumerableSet.AddressSet;
 *
 *     # Declare a set state variable
 *     EnumerableSet.AddressSet private mySet;
 * }
 * ```
 *
 * As of v3.3.0, sets of type `bytes32` (`Bytes32Set`), `address` (`AddressSet`)
 * and `uint256` (`UintSet`) are supported.
'''
class EnumerableSet:
    # To implement this library for multiple types with as little code
    # repetition as possible, we write it in terms of a generic Set type with
    # bytes32 values.
    # The Set implementation uses private functions, and user-facing
    # implementations (such as AddressSet) are just wrappers around the
    # underlying Set.
    # This means that we can only create new EnumerableSets for types that fit
    # in bytes32.

    def __init__(self):
        self.set = set()

    '''
     * @dev Add a value to a set. O(1).
     *
     * Returns true if the value was added to the set, that is if it was not
     * already present.
    '''
    def add(self, value):
        if (not self.contains(value)):
            self.set.add(value);
            return True;
        else:
            return False;

    '''
     * @dev Removes a value from a set. O(1).
     *
     * Returns true if the value was removed from the set, that is if it was
     * present.
    '''
    def remove(self, value):
        if (self.contains(value)):
            self.set.remove(value);
            return True;
        else:
            return False;

    '''
     * @dev Returns true if the value is in the set. O(1).
    '''
    def contains(self, value):
        return value in self.set;

    '''
     * @dev Returns the number of values on the set. O(1).
    '''
    def length(self):
        return len(self.set);

    '''
     * @dev Returns the value stored at position `index` in the set. O(1).
     *
     * Note that there are no guarantees on the ordering of values inside the
     * array, and it may change when more values are added or removed.
     *
     * Requirements:
     *
     * - `index` must be strictly less than {length}.
    '''
    def at(self, index):
        return list(self.set)[index];

    '''
     * @dev Return the entire set in an array
     *
     * WARNING: This operation will copy the entire storage to memory, which can be quite expensive. This is designed
     * to mostly be used by view accessors that are queried without any gas fees. Developers should keep in mind that
     * this function has an unbounded cost, and using it as part of a state-changing function may render the function
     * uncallable if the set grows to a point where copying to memory consumes too much gas to fit in a block.
    '''
    def values(self):
        return list(self.set);
